/*
 * timout.c	Advanced timeout handling for file system calls
 *		to avoid deadlocks on remote file shares.
 *
 * Version:	0.2 11-Dec-2012 Fink
 *
 * Copyright 2011,2012 Werner Fink, 2011,2012 SUSE LINUX Products GmbH, Germany.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Author:	Werner Fink <werner@suse.de>, 2011
 */

#ifndef _GNU_SOURCE
# define _GNU_SOURCE
#endif

#ifdef _FEATURES_H
# error Include local config.h before any system header file
#endif
#include "config.h"

#ifndef WITH_TIMEOUT_STAT
# define WITH_TIMEOUT_STAT	0
#endif

#include <errno.h>
#include <pthread.h>
#include <setjmp.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <unistd.h>
#include <wait.h>

#include "timeout.h"
#if defined(HAVE_SYSCALL_STATX) || (defined(HAVE_DECL_SYS_STATX) && HAVE_DECL_SYS_STATX == 1)
#include "statx.h"
#endif

#if !defined(__STDC_VERSION__) || (__STDC_VERSION__ < 199901L)
# ifndef  destructor
#  define destructor		__destructor__
# endif
# ifndef  constructor
#  define constructor		__constructor__
# endif
# ifndef  packed
#  define packed		__packed__
# endif
# ifndef  inline
#  define inline		__inline__
# endif
# ifndef  unused
#  define unused		__unused__
# endif
# ifndef  volatile
#  define volatile		__volatile__
# endif
#endif
#ifndef  attribute
# define attribute(attr)	__attribute__(attr)
#endif

#if defined __GNUC__
# undef strcpy
# define strcpy(d,s)		__builtin_strcpy((d),(s))   /* Without boundary check please */
#endif

#if !defined(HAVE_SYSCALL_STATX) && !(defined(HAVE_DECL_SYS_STATX) && HAVE_DECL_SYS_STATX == 1)
#if WITH_TIMEOUT_STAT
static sigjmp_buf jenv;
static void sigjump(int sig attribute((unused)))
{
	siglongjmp(jenv, 1);
}
#endif
#endif

#if WITH_TIMEOUT_STAT == 2
/*
 * The structure used for communication between the processes
 */
typedef struct _handle {
	pthread_mutex_t mutex;
	pthread_cond_t waiton, result;
	volatile int todo;
	pid_t active;
	int errcode;
	struct stat argument;
	stat_t function;
	size_t len;
	char path[0];
} attribute((packed)) handle_t;

/*
 * Using a forked process for doing e.g. stat(2) system call as this
 * allows us to send e.g. SIGKILL to this process if it hangs in `D'
 * state on a file share due a stalled NFS server.  This does not work
 * with (p)threads as SIGKILL would kill all threads including main.
 */

#if !defined(HAVE_SYSCALL_STATX) && !(defined(HAVE_DECL_SYS_STATX) && HAVE_DECL_SYS_STATX == 1)
static volatile pid_t active;
static handle_t *handle;
static const size_t buflen = PATH_MAX+sizeof(handle_t)+1;

static void sigchild(int sig attribute((unused)))
{
	pid_t pid = waitpid(active, NULL, WNOHANG|WUNTRACED);
	if (pid <= 0)
		return;
	if (errno == ECHILD)
		return;
	active = 0;
}

static void attribute((constructor)) start(void)
{
	sigset_t sigset, oldset;
	struct sigaction act;
	pthread_mutexattr_t mattr;
	pthread_condattr_t cattr;

	memset(&act, 0, sizeof(act));
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	act.sa_handler = sigchild;
	sigaction(SIGCHLD, &act, 0);

	if (!handle) {
		handle = mmap(NULL, buflen, PROT_READ|PROT_WRITE,
			      MAP_ANONYMOUS|MAP_SHARED, -1, 0);
		if (handle == MAP_FAILED)
			goto error;
		memset(handle, 0, buflen);

		if (pthread_mutexattr_init(&mattr) < 0)
			goto error;
		if (pthread_mutexattr_setpshared(&mattr, PTHREAD_PROCESS_SHARED) < 0)
			goto error;
		if (pthread_mutex_init(&handle->mutex, &mattr) < 0)
			goto error;
		pthread_mutexattr_destroy(&mattr);

		if (pthread_condattr_init(&cattr) < 0)
			goto error;
		if (pthread_condattr_setpshared(&cattr, PTHREAD_PROCESS_SHARED) < 0)
			goto error;
		if (pthread_cond_init(&handle->waiton, &cattr) < 0)
			goto error;
		if (pthread_cond_init(&handle->result, &cattr) < 0)
			goto error;
		pthread_condattr_destroy(&cattr);
	}

	handle->todo = 0;
	if ((active = fork()) < 0)
		goto error;

	if (active) {
		handle->active = active;
		return;
	}
	sigemptyset(&sigset);
	sigaddset(&sigset, SIGALRM);
	sigprocmask(SIG_BLOCK, &sigset, &oldset);

	act.sa_handler = SIG_DFL;
	sigaction(SIGCHLD, &act, 0);

	do {
		pthread_mutex_lock(&handle->mutex);
		while (!handle->todo)
			pthread_cond_wait(&handle->waiton, &handle->mutex);

		if (handle->function(handle->path, &handle->argument) < 0)
			handle->errcode = errno;

		handle->todo = 0;
		pthread_cond_broadcast(&handle->result);
		pthread_mutex_unlock(&handle->mutex);

	} while (handle->active);

	sigprocmask(SIG_SETMASK, &oldset, NULL);
	exit(0);
error:
	if (handle && handle != MAP_FAILED)
		munmap(handle, buflen);
	handle = NULL;
}

static void attribute((destructor)) stop(void)
{
	if (active && waitpid(active, NULL, WNOHANG|WUNTRACED) == 0)
		kill(active, SIGKILL);
}
#endif

/*
 * External routine
 *
 * Execute stat(2) system call with timeout to avoid deadlock
 * on network based file systems.
 *
 */
int
timeout(stat_t function, const char *path, struct stat *restrict argument, time_t seconds)
{
#if !defined(HAVE_SYSCALL_STATX) && !(defined(HAVE_DECL_SYS_STATX) && HAVE_DECL_SYS_STATX == 1)
	struct sigaction alrm_act, new_act;
	sigset_t sigset, oldset;

	if (active <= 0)	/* Oops, last one failed therefore clear status and restart */
		start();
	if (!handle)		/* No shared memory area */
#endif
		return function(path, argument);
#if !defined(HAVE_SYSCALL_STATX) && !(defined(HAVE_DECL_SYS_STATX) && HAVE_DECL_SYS_STATX == 1)
	handle->len = strlen(path) + 1;
	if (handle->len >= PATH_MAX) {
		errno = ENAMETOOLONG;
		goto error;
	}
	handle->errcode = 0;
	handle->argument = *argument;
	handle->function = function;
	strcpy(handle->path, path);

	sigemptyset(&sigset);
	sigaddset(&sigset, SIGALRM);
	sigaddset(&sigset, SIGINT);
	sigprocmask(SIG_UNBLOCK, &sigset, &oldset);

	memset(&new_act, 0, sizeof(new_act));
	sigemptyset(&new_act.sa_mask);
	new_act.sa_flags = SA_RESETHAND;
	if (sigsetjmp(jenv, 1))
		goto timed;

	new_act.sa_handler = sigjump;
	sigaction(SIGALRM, &new_act, &alrm_act);
	alarm(seconds);

	pthread_mutex_lock(&handle->mutex);
	handle->todo = 1;
	pthread_cond_broadcast(&handle->waiton);
	while (handle->todo)
		pthread_cond_wait(&handle->result, &handle->mutex);
	pthread_mutex_unlock(&handle->mutex);

	alarm(0);
	sigaction(SIGALRM, &alrm_act, NULL);

	if (handle->errcode) {
		errno = handle->errcode;
		goto error;
	}

	*argument = handle->argument;
	sigprocmask(SIG_SETMASK, &oldset, NULL);

	return 0;
timed:
	(void) alarm(0);
	sigaction(SIGALRM, &alrm_act, NULL);
	sigprocmask(SIG_SETMASK, &oldset, NULL);
	stop();
	errno = ETIMEDOUT;
error:
	return -1;
#endif
}
#elif WITH_TIMEOUT_STAT == 1
/*
 * External routine
 *
 * Execute stat(2) system call with timeout to avoid deadlock
 * on network based file systems.
 *
 */
int
timeout(stat_t function, const char *path, struct stat *restrict argument, time_t seconds)
{
	struct sigaction alrm_act, pipe_act, new_act;
	sigset_t sigset, oldset;
	int ret = 0, pipes[4];
	pid_t pid = 0;
	ssize_t len;

	if (pipe(&pipes[0]) < 0)
		goto error;
	switch ((pid = fork())) {
	case -1:
		close(pipes[0]);
		close(pipes[1]);
		goto error;
	case 0:
		new_act.sa_handler = SIG_DFL;
		sigaction(SIGALRM, &new_act, NULL);
		close(pipes[0]);
		if ((ret = function(path, argument)) == 0)
			do
				len = write(pipes[1], argument, sizeof(struct stat));
			while (len < 0 && errno == EINTR);
		close(pipes[1]);
		exit(ret);
	default:
		close(pipes[1]);

		sigemptyset(&sigset);
		sigaddset(&sigset, SIGALRM);
		sigaddset(&sigset, SIGPIPE);
		sigprocmask(SIG_UNBLOCK, &sigset, &oldset);

		memset(&new_act, 0, sizeof(new_act));
		sigemptyset(&new_act.sa_mask);

		if (sigsetjmp(jenv, 1))
			goto timed;

		new_act.sa_handler = sigjump;
		sigaction(SIGALRM, &new_act, &alrm_act);
		sigaction(SIGPIPE, &new_act, &pipe_act);
		alarm(seconds);
		if (read(pipes[0], argument, sizeof(struct stat)) == 0) {
			errno = EFAULT;
			ret = -1;
		}
		(void)alarm(0);
		sigaction(SIGPIPE, &pipe_act, NULL);
		sigaction(SIGALRM, &alrm_act, NULL);

		close(pipes[0]);
		waitpid(pid, NULL, 0);
		break;
	}
	return ret;
timed:
	(void)alarm(0);
	sigaction(SIGPIPE, &pipe_act, NULL);
	sigaction(SIGALRM, &alrm_act, NULL);
	if (waitpid(0, NULL, WNOHANG) == 0)
		kill(pid, SIGKILL);
	errno = ETIMEDOUT;
error:
	return -1;
}
#endif

/*
 * End of timeout.c
 */
